import scala.util.{Try,Failure,Success}

sealed trait Expr
case class Number(value: Int) extends Expr
case class Variable(val name: String) extends Expr
case class BinOp(operator: Char, left: Expr, right: Expr) extends Expr

sealed trait Order
case object LessThan extends Order
case object Equals extends Order
case object GreaterThan extends Order

trait Ordered[T] { self: T =>
  def compare(other: T): Order
  def >(other: T) =
    self.compare(other) == GreaterThan
  def <(other: T) =
    self.compare(other) == LessThan
  def <=(other: T) =
    self.compare(other) != GreaterThan
}

class Rational(val numer: Int, val denom: Int) extends Ordered[Rational] {
  def compare(other: Rational) =
    if (this.numer > other.numer) GreaterThan
    else if (this.numer == other.numer) Equals
    else LessThan
}

object Expr {
  def main(args: Array[String]): Unit = {
    for (i <- 1 to 10) yield i * 2
    val expr = BinOp(
      '+',
      Number(4),
      Variable("bar"))
    implicit val memory: String => Option[Int] = Map("bar" -> 7).get
    println(eval(expr))
    val arr = Array(1,2,3,4,5)
    val x: Try[Int] = Try(arr(10))
    x match {
      case Failure(e) =>
        println("did not work: " + e.getClass.getName)
      case Success(v) =>
        println("it did work: " + v)
    }
    def asLongAs(c: => Boolean)(body: => Unit): Unit =
      if (c) { body; asLongAs(c)(body) } else ()
    var i = 0
    asLongAs(i < 10)(i += 1)
    println(i)
  }

  def combine(a: Option[Int], b: Option[Int], f: (Int, Int) => Int): Option[Int] = for {
    a <- a
    b <- b
  } yield f(a,b)

  def eval(expr: Expr)(implicit memory: String => Option[Int]): Option[Int] = expr match {
    case Number(value) => Some(value)
    case Variable(name) => memory(name)
    case BinOp('+',l,r) => combine(eval(l),eval(r), (a,b) => a + b)
    case BinOp('-',l,r) => combine(eval(l),eval(r), (a,b) => a - b)
    case BinOp('*',l,r) => combine(eval(l),eval(r), (a,b) => a * b)
    case BinOp('/',l,r) => combine(eval(l),eval(r), (a,b) => a / b)
  }


}



